#include "Error.h"
#include "Game.h"
#include "Input.h"
#include "Renderer.h"
#include "Single.h"
#include "Sound.h"
#include "Tunnel.h"

#include <stdlib.h>
#include <time.h>

#define FPS (60)
#define MS_PER_FRAME (1000 / FPS)
#define SPF ((float)MS_PER_FRAME / 1000.f)

LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
	case WM_ACTIVATE:
		if (Input::Try_Get() != 0)
		{
			Input::Get()->Acquire();
		}
		return 0;

	case WM_DESTROY:
        PostQuitMessage(0);
        return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

INT WINAPI wWinMain( HINSTANCE hInst, HINSTANCE, LPWSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc =
    {
        sizeof( WNDCLASSEX ), CS_CLASSDC, MsgProc, 0L, 0L,
        GetModuleHandle( NULL ), NULL, NULL, NULL, NULL,
        L"Cavity Explorer", NULL
    };
    RegisterClassEx(&wc);

    // Create the application's window
    HWND hwnd = 
		CreateWindow(
			L"Cavity Explorer",
			L"Cavity Explorer",
            WS_OVERLAPPEDWINDOW, 
			100, 100, 640, 480,
            NULL, NULL, wc.hInstance, NULL);

    // Show the window
    ShowWindow(hwnd, SW_SHOWDEFAULT);
    UpdateWindow(hwnd);

	// Init the random seed
	srand(time(NULL));

	// Set the timer accuracy
	timeBeginPeriod(1);

	try
	{
		// Create the renderer
		Single<Input> input = Make_Single(new Input(hInst, hwnd));
		Single<Renderer> renderer = Make_Single(new Renderer(hwnd));
		Single<Sound> sound = Make_Single(new Sound());
		Single<Game> game = Make_Single(new Game());

        // Enter the main loop
		bool finished = false;
		DWORD time_last = timeGetTime();
        while (!finished)
        {
			// Handle Windows messages
	        MSG msg;
            while (PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
				if (msg.message == WM_QUIT)
					finished = true;
            }

			// Update the logic
			DWORD time_now = timeGetTime();
			if ((time_now < time_last) ||
				((time_now - time_last) > 1000))
			{ // Timer wrap/skip
				time_last = time_now;
				continue;
			}

			bool has_updated = false;
			while ((time_now - time_last) >= MS_PER_FRAME)
			{
				input->Poll();
				sound->Update(SPF);
				game->Update(SPF, input.Get());

				time_last += MS_PER_FRAME;
				has_updated = true;
			}

			if (!has_updated)
				continue;

			// Render the screen
			renderer->Clear();
			renderer->Begin_Scene();
			game->Render(renderer.Get());
			renderer->End_Scene();
			renderer->Flip();
        }
	}
	catch (Error e)
	{
		MessageBoxA(hwnd, e.Get_Msg(), "Error!", MB_OK | MB_ICONERROR | MB_SYSTEMMODAL);
	}

	// Restore the timer accuracy
	timeEndPeriod(1);

    return 0;
}
