#include "Error.h"
#include "Renderer.h"
#include "Vertex.h"

Renderer::Renderer(HWND wnd)
	: m_d3d(0)
	, m_d3d_device(0)
	, m_effect(0)
	, m_font_big(0)
	, m_font_small(0)
{
	D3DXMatrixIdentity(&m_mtx_world);
	D3DXMatrixIdentity(&m_mtx_view);
	D3DXMatrixIdentity(&m_mtx_proj);

    // Create the D3D object.
	m_d3d = Direct3DCreate9(D3D_SDK_VERSION);
    if (m_d3d == 0)
        throw Error("Error creating the Direct3D object!");

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    // Create the D3DDevice
	HRESULT res =
		m_d3d->CreateDevice( 
			D3DADAPTER_DEFAULT, 
			D3DDEVTYPE_HAL,
			wnd,
			D3DCREATE_MIXED_VERTEXPROCESSING,
            &d3dpp,
			&m_d3d_device);
    if (FAILED(res))
		throw Error("Error creating the Direct3D device!");

	// Get the viewport to get the resolution
	D3DVIEWPORT9 viewport;
	m_d3d_device->GetViewport(&viewport);
	m_width = (int)viewport.Width;
	m_height = (int)viewport.Height;


	// Set the device options
    m_d3d_device->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    m_d3d_device->SetRenderState(D3DRS_LIGHTING, FALSE);
    m_d3d_device->SetRenderState(D3DRS_ZENABLE, TRUE);

	// Create the effect
	res = D3DXCreateEffectFromFile(m_d3d_device, L"Gfx.fx", NULL, NULL, D3DXFX_NOT_CLONEABLE, NULL, &m_effect, NULL);
	if (FAILED(res))
		throw Error("Error loading the shaders from 'Gfx.fx'!");

	// Create the fonts
	res = D3DXCreateFont(
		m_d3d_device,
		45, 0, 
		FW_BOLD, 1, FALSE, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
		L"Arial", &m_font_big);
	if (FAILED(res))
		throw Error("Error initializing the font (big)!");

	res = D3DXCreateFont(
		m_d3d_device,
		18, 0, 
		FW_BOLD, 1, FALSE, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
		L"Arial", &m_font_small);
	if (FAILED(res))
		throw Error("Error initializing the font (small)!");
}

Renderer::~Renderer()
{
	if (m_font_small != 0)
		m_font_small->Release();

	if (m_font_big != 0)
		m_font_big->Release();

	if (m_effect != 0)
		m_effect->Release();

	if (m_d3d_device != 0)
		m_d3d_device->Release();

	if (m_d3d != 0)
		m_d3d->Release();
}

int Renderer::Get_Width() const
{
	return m_width;
}

int Renderer::Get_Height() const
{
	return m_height;
}

void Renderer::Clear()
{
	m_d3d_device->Clear(
		0, 
		NULL, 
		D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
		D3DCOLOR_XRGB(0, 0, 0),
		1.f,
		0);
}

void Renderer::Begin_Scene()
{
	m_d3d_device->BeginScene();
}

void Renderer::Set_Param(const char *param_str, float param_val)
{
	m_effect->SetFloat(param_str, param_val);
}

void Renderer::Render_Mesh(const Mesh *mesh)
{
	if (mesh == 0)
		return;

	int vcount = mesh->Get_Size();
	int prim_count = vcount - 2;
	if (prim_count <= 0)
		return;

	// Set the vertex format
	m_d3d_device->SetFVF(VERTEX_FVF);

	// Get the technique
	D3DXHANDLE tech = m_effect->GetTechniqueByName("Basic");
	if (tech == 0)
		throw Error("Rendering technique not found!");
	m_effect->SetTechnique(tech);

	// Set the effect matrices
	D3DXMATRIXA16 world_mtx = m_mtx_world * m_mtx_view;
	m_effect->SetMatrix("mtx_world", &world_mtx);
	D3DXMATRIXA16 world_mtx_inv;
	D3DXMatrixInverse(&world_mtx_inv, NULL, &world_mtx);
	m_effect->SetMatrix("mtx_world_inv", &world_mtx_inv);
	m_effect->SetMatrix("mtx_proj", &m_mtx_proj);

	// Render all the passes
	UINT passes;
	m_effect->Begin(&passes, 0);
	for (UINT pass = 0; pass < passes; ++pass)
	{
		m_effect->BeginPass(pass);

		int vcount = mesh->Get_Size();
		m_d3d_device->DrawPrimitiveUP(
			D3DPT_TRIANGLESTRIP,
			prim_count,
			mesh->Get_Vertices(),
			sizeof(Vertex));

		m_effect->EndPass();
	}
	m_effect->End();
}

void Renderer::Write_Text(int x, int y, LPCTSTR str, D3DCOLOR color, bool center, bool big_font)
{
	DWORD a = (color >> 24) & 0xff;
	Do_Write_Text(x + 2, y + 2, str, D3DCOLOR_ARGB(a, 0, 0, 0), center, big_font);
	Do_Write_Text(x, y, str, color, center, big_font);
}

void Renderer::End_Scene()
{
	m_d3d_device->EndScene();
}

void Renderer::Flip()
{
	m_d3d_device->Present(NULL, NULL, NULL, NULL);
}

void Renderer::Set_World_Matrix(const D3DXMATRIX &mtx)
{
	m_mtx_world = mtx;
}

void Renderer::Set_View_Matrix(const D3DXMATRIX &mtx)
{
	m_mtx_view = mtx;
}

void Renderer::Set_Projection_Matrix()
{
    D3DXMatrixPerspectiveFovLH(
		&m_mtx_proj, 
		D3DX_PI * 0.25f,
		(float)m_width / (float)m_height,
		0.1f, 250.0f);
}

void Renderer::Do_Write_Text(int x, int y, LPCTSTR str, D3DCOLOR color, bool center, bool big_font)
{
	ID3DXFont *font = big_font ? m_font_big : m_font_small;

	RECT rect;
	rect.left = x;
	rect.right = x;
	rect.top = y;
	rect.bottom = y;

	font->DrawText(
		NULL,
		str,
		-1,
		&rect,
		DT_SINGLELINE | DT_NOCLIP | (center ? DT_CENTER : 0),
		color);
}
